package de.tu_darmstadt.tk.SmartHomeNetworkSim.view;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.util.Collection;
import java.util.Observable;
import java.util.Observer;

import javax.swing.JPanel;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.Controller;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Connection;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Model;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Port;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.SmartDevice;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.protocols.MQTT_protocol;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.view.VisualisationInteractor;

/**
 * Panel which visualizes the SmartHome Network, it's Devices and Connections
 * 
 *
 * @author Andreas T. Meyer-Berg
 */
@SuppressWarnings("serial")
public class VisualisationPanel extends JPanel implements Observer {

	/**
	 * Smart Home model which is visualized
	 */
	private Model model;

	/**
	 * Controller to notify when the model changes
	 */
	private Controller control;

	/**
	 * Listener which processes the GUI Interactions
	 */
	private VisualisationInteractor interactor;
	
	/**
	 * Initializes the Visualization Panel
	 * 
	 * @param model
	 *            Model to visualize
	 * @param control
	 *            Control, which changes the model
	 */
	public VisualisationPanel(Model model, Controller control) {
		super();

		this.model = model;
		this.control = control;
		
		this.interactor = new VisualisationInteractor(model, control, this);
		this.addMouseMotionListener(interactor);
		this.addMouseListener(interactor);
		this.addKeyListener(interactor);
	}

	/**
	 * Performs further initializations, which have to be performed, after the
	 * model was made visible
	 */
	public void delayedInit() {
		this.addComponentListener(new ComponentListener() {

			@Override
			public void componentShown(ComponentEvent e) {
			}

			@Override
			public void componentResized(ComponentEvent e) {
				control.setDimension(getWidth(), getHeight(), model.getDepth(),
						true);
				repaint();
			}

			@Override
			public void componentMoved(ComponentEvent e) {
			}

			@Override
			public void componentHidden(ComponentEvent e) {
			}
		});
		control.setDimension(getWidth(), getHeight(), model.getDepth(),
				true);
		repaint();
	}

	@Override
	public void paint(Graphics g) {
		// paint white background
		g.setColor(Color.white);
		g.fillRect(0, 0, this.getWidth(), this.getHeight());

		paintConnections(g);

		paintDevices(g);
		
		paintDrag(g);
		
		g.setColor(Color.RED);
		String modus;
		switch (interactor.mode) {
		case VisualisationInteractor.DRAG_CONNECTION:
			modus="Drag Connection";
			break;
		case VisualisationInteractor.SELECTED_DRAG:
			modus="Drag Device";
			break;
		case VisualisationInteractor.DRAG_SELECT:
			modus="Drag Select";
			break;
		case VisualisationInteractor.NOTHING:
			modus="Nothing";
			break;
		case VisualisationInteractor.RIGHTCLICK_MENU:
			modus="RightClick Menu";
			break;
		case VisualisationInteractor.SELECTED:
			modus="Selected";
			break;
		default:
			modus="Uknown";
			break;
		}
		g.drawString("Debug(Modus:"+modus+")", 0, 10);
		if(interactor.controlDown)
			g.drawString("Control down", 0, 20);
	}

	/**
	 * Paints the smart devices of the Model
	 * 
	 * @param g
	 *            the Graphics object to visualize on
	 */
	public void paintDevices(Graphics g) {
		//Values for dragging of multiple Devices
		int x_offset = 0;
		int y_offset = 0;
		if(interactor.mode == VisualisationInteractor.SELECTED_DRAG && !interactor.selectedDevices.isEmpty()){
			x_offset = interactor.dragged_x-interactor.dragged.getX();
			y_offset = interactor.dragged_y-interactor.dragged.getY();
		}
		//Paint Devices
		for (SmartDevice s: model.getDevices()) {
			int x = s.getX();
			int y = s.getY();
			if(interactor.mode==VisualisationInteractor.SELECTED_DRAG && interactor.selectedDevices.contains(s)) {
				// Update visualization of dragged object
				x += x_offset;
				y += y_offset;
			}
			if(//(interactor.mode==VisualisationInteractor.SELECTED||interactor.mode==VisualisationInteractor.DRAG_SELECT ||interactor.mode == VisualisationInteractor.SELECTED_DRAG || interactor.mode == VisualisationInteractor.RIGHTCLICK_MENU)&&
					(interactor.selectedDevices.contains(s)^interactor.selectedDevicesDrag.contains(s))){
				//HighlightSelected Devices
				g.setColor(new Color(135,206,235));
			}else
				g.setColor(Color.WHITE);
			g.fillOval(x - control.getDevice_visualization_radius(), y - control.getDevice_visualization_radius(),
					2 * control.getDevice_visualization_radius() - 1, 2 * control.getDevice_visualization_radius() - 1);
			g.setColor(Color.BLACK);
			g.drawOval(x - control.getDevice_visualization_radius(), y - control.getDevice_visualization_radius(),
					2 * control.getDevice_visualization_radius() - 1, 2 * control.getDevice_visualization_radius() - 1);
			g.setColor(Color.BLUE);
			g.drawOval(x - control.getDevice_visualization_radius() + 2,
					y - control.getDevice_visualization_radius() + 2, 2 * control.getDevice_visualization_radius() - 5,
					2 * control.getDevice_visualization_radius() - 5);
			g.setColor(Color.BLACK);
			g.drawString(s.getName(),
					x - g.getFontMetrics().stringWidth(s.getName()) / 2, y
							+ control.getDevice_visualization_radius() + 11);
		}
	}

	/**
	 * Paints the Connections
	 * 
	 * @param g
	 *            the Graphics object to visualize on
	 */
	private void paintConnections(Graphics g) {
		//Values for dragging of multiple Devices
		int x_offset = 0;
		int y_offset = 0;
		if(interactor.mode == VisualisationInteractor.SELECTED_DRAG && !interactor.selectedDevices.isEmpty()){
			x_offset = interactor.dragged_x-interactor.dragged.getX();
			y_offset = interactor.dragged_y-interactor.dragged.getY();
		}
		// For all Connections
		for (Connection c : model.getConnections()) {
			Color connectionState;
			switch (c.getStatus()) {
			case Connection.ACTIVE:
				connectionState = Color.GREEN;
				break;
			case Connection.HALTED:
				connectionState = Color.ORANGE;
				break;
			case Connection.DONE:
				connectionState = Color.GRAY;
				break;
			case Connection.TERMINATED:
			case Connection.FINISHED:
				connectionState = Color.RED;
				break;
			default:
				connectionState = Color.BLUE;
				break;
			}
			//g.setColor(connectionState);
			/**
			 * All Devices that are part of the connection
			 */
			Collection<Port> d = c.getParticipants();
			if(d.size()==0){
				System.out.println("WARNING: Empty, undeleted Connection:"+c.toString());
				continue;
			}
			if(c.getProtocol() instanceof MQTT_protocol){
				Port broker = c.getProtocol().getDevicesWithRole(0).iterator().next();
				int broker_x = 0, broker_y = 0;
				if(broker==null){
					/**
					 * if broker was removed, paint broker at the average position
					 */
					for(Port p: c.getParticipants()){
						if(interactor.mode==VisualisationInteractor.SELECTED_DRAG && interactor.selectedDevices.contains(p.getOwner())){
							broker_x+=p.getOwner().getX()+x_offset;
							broker_y+=p.getOwner().getY()+y_offset;
						}else{

							broker_x+=p.getOwner().getX();
							broker_y+=p.getOwner().getY();
						}
					}
					broker_x/=c.getParticipants().size();
					broker_y/=c.getParticipants().size();
					g.fillOval(broker_x-control.getDevice_visualization_radius()/4, broker_y-control.getDevice_visualization_radius()/4, control.getDevice_visualization_radius()/2, control.getDevice_visualization_radius()/2);
				}else{
					if(interactor.mode==VisualisationInteractor.SELECTED_DRAG && interactor.selectedDevices.contains(broker.getOwner())){
						broker_x=broker.getOwner().getX()+x_offset;
						broker_y=broker.getOwner().getY()+y_offset;						
					}else{
						broker_x=broker.getOwner().getX();
						broker_y=broker.getOwner().getY();
					}
				}
				for(Port sd:d){
					if(!model.getDevices().contains(sd.getOwner())){
						SmartDevice removed = sd.getOwner();
						g.setColor(Color.RED);
						g.fillOval(removed.getX()-control.getDevice_visualization_radius()/4, removed.getY()-control.getDevice_visualization_radius()/4, control.getDevice_visualization_radius()/2, control.getDevice_visualization_radius()/2);
					}else{
						g.setColor(connectionState);
					}
					
					if (broker!= sd) {
						// Check if dragged object
						int s_x, s_y, sd_x, sd_y;
						if (interactor.mode==VisualisationInteractor.SELECTED_DRAG && broker != null && interactor.selectedDevices.contains(broker.getOwner())) {
							s_x = broker.getOwner().getX()+x_offset;
							s_y = broker.getOwner().getY()+y_offset;
						} else if(broker==null){
							/**
							 * if broker was removed, paint edges to the previously calculated average position
							 */
							s_x = broker_x;
							s_y = broker_y;
						}else{
							s_x = broker_x;
							s_y = broker_y;
						}
						if (interactor.mode==VisualisationInteractor.SELECTED_DRAG && interactor.selectedDevices.contains(sd.getOwner())) {
							sd_x = sd.getOwner().getX()+x_offset;
							sd_y = sd.getOwner().getY()+y_offset;
						} else {
							sd_x = sd.getOwner().getX();
							sd_y = sd.getOwner().getY();
						}
						g.drawLine(s_x, s_y, sd_x, sd_y);
					}
				}
			}
			else if (d.size() == 2) {
				Port s = c.getSource();
				for (Port sd : d) {
					if(!model.getDevices().contains(sd.getOwner())){
						SmartDevice removed = sd.getOwner();
						g.setColor(Color.RED);
						g.fillOval(removed.getX()-control.getDevice_visualization_radius()/4, removed.getY()-control.getDevice_visualization_radius()/4, control.getDevice_visualization_radius()/2, control.getDevice_visualization_radius()/2);
					}else{
						g.setColor(connectionState);
					}
					if (s != sd) {
						// Check if dragged object
						int s_x, s_y, sd_x, sd_y;
						if (interactor.mode==VisualisationInteractor.SELECTED_DRAG && interactor.selectedDevices.contains(s.getOwner())) {
							s_x = s.getOwner().getX()+x_offset;
							s_y = s.getOwner().getY()+y_offset;
						} else {
							s_x = s.getOwner().getX();
							s_y = s.getOwner().getY();
						}
						if (interactor.mode==VisualisationInteractor.SELECTED_DRAG && interactor.selectedDevices.contains(sd.getOwner())) {
							sd_x = sd.getOwner().getX()+x_offset;
							sd_y = sd.getOwner().getY()+y_offset;
						} else {
							sd_x = sd.getOwner().getX();
							sd_y = sd.getOwner().getY();
						}
						g.drawLine(s_x, s_y, sd_x, sd_y);
					}
				}
			} else if (d.size() == 1) {

			} else if (d.size() <= 0) {
				// Invalid Connection
			} else {
				// Draw MultiConnection
			}

		}
		// paint new in progress connection, if a connection or link is in
		// creation
		if (interactor.mode == VisualisationInteractor.DRAG_CONNECTION && interactor.connectionFrom != null){
			g.setColor(Color.ORANGE);
			g.drawLine(interactor.connectionFrom.getX(),
					interactor.connectionFrom.getY(), interactor.dragged_x,
					interactor.dragged_y);
		}
	}
	
	/**
	 * Paints the 
	 * @param g
	 */
	public void paintDrag(Graphics g){
		if(interactor.mode!=VisualisationInteractor.DRAG_SELECT)
			return;	
		int low_x = Math.min(interactor.dragged_x, interactor.dragged_x_start);
		int low_y = Math.min(interactor.dragged_y, interactor.dragged_y_start);
		int high_x = Math.max(interactor.dragged_x, interactor.dragged_x_start);
		int high_y = Math.max(interactor.dragged_y, interactor.dragged_y_start);
		g.setColor(Color.BLUE);
		g.drawRect(low_x, low_y, high_x-low_x, high_y-low_y);

	}
	@Override
	public void update(Observable o, Object arg) {
		repaint();
	}

}
